from abc import ABC, abstractmethod
from typing import List

from fastapi import APIRouter, Depends

from jupyverse_api import Router

from ..app import App
from ..auth import Auth, User
from .models import Terminal


class Terminals(Router, ABC):
    def __init__(self, app: App, auth: Auth):
        super().__init__(app=app)

        router = APIRouter()

        @router.get("/api/terminals")
        async def get_terminals(
            user: User = Depends(auth.current_user({"terminals": ["read"]})),
        ) -> List[Terminal]:
            return await self.get_terminals(user)

        @router.post("/api/terminals")
        async def create_terminal(
            user: User = Depends(auth.current_user({"terminals": ["write"]})),
        ):
            return await self.create_terminal(user)

        @router.delete("/api/terminals/{name}", status_code=204)
        async def delete_terminal(
            name: str,
            user: User = Depends(auth.current_user(permissions={"terminals": ["write"]})),
        ):
            return await self.delete_terminal(name, user)

        @router.websocket("/terminals/websocket/{name}")
        async def terminal_websocket(
            name,
            websocket_permissions=Depends(
                auth.websocket_auth(permissions={"terminals": ["read", "execute"]})
            ),
        ):
            return await self.terminal_websocket(name, websocket_permissions)

        self.include_router(router)

    @abstractmethod
    async def get_terminals(
        self,
        user: User,
    ) -> List[Terminal]:
        ...

    @abstractmethod
    async def create_terminal(
        self,
        user: User,
    ):
        ...

    @abstractmethod
    async def delete_terminal(
        self,
        name: str,
        user: User,
    ):
        ...

    @abstractmethod
    async def terminal_websocket(
        self,
        name,
        websocket_permissions,
    ):
        ...


class TerminalServer(ABC):
    @abstractmethod
    async def serve(self, websocket, permissions):
        ...

    @abstractmethod
    def quit(self, websocket):
        ...



from asphalt.core import Component, Context

from jupyverse_api.app import App
from jupyverse_api.auth import Auth, AuthConfig

from .config import _AuthFiefConfig
from .routes import auth_factory


class AuthFiefComponent(Component):
    def __init__(self, **kwargs):
        self.auth_fief_config = _AuthFiefConfig(**kwargs)

    async def start(
        self,
        ctx: Context,
    ) -> None:
        ctx.add_resource(self.auth_fief_config, types=AuthConfig)

        app = await ctx.request_resource(App)

        auth_fief = auth_factory(app, self.auth_fief_config)
        ctx.add_resource(auth_fief, types=Auth)



from __future__ import annotations

import asyncio
import json
import os
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import httpx
from fastapi import APIRouter, Cookie, Depends, HTTPException, Request, WebSocket, status
from fastapi.responses import RedirectResponse
from jupyterhub.services.auth import HubOAuth
from jupyterhub.utils import isoformat
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select
from typing_extensions import Annotated

from jupyverse_api import Router
from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User

from .db import UserDB
from .models import JupyterHubUser


def auth_factory(
    app: App,
    db_session: AsyncSession,
    http_client: httpx.AsyncClient,
):
    class AuthJupyterHub(Auth, Router):
        def __init__(self) -> None:
            super().__init__(app)
            self.hub_auth = HubOAuth()
            self.db_lock = asyncio.Lock()
            self.activity_url = os.environ.get("JUPYTERHUB_ACTIVITY_URL")
            self.server_name = os.environ.get("JUPYTERHUB_SERVER_NAME")
            self.background_tasks = set()

            router = APIRouter()

            @router.get("/oauth_callback")
            async def get_oauth_callback(
                request: Request,
                code: str | None = None,
                state: str | None = None,
            ):
                if code is None:
                    raise HTTPException(status_code=status.HTTP_403_FORBIDDEN)

                cookie_state = request.cookies.get(self.hub_auth.state_cookie_name)
                if state is None or state != cookie_state:
                    raise HTTPException(status_code=status.HTTP_403_FORBIDDEN)

                token = self.hub_auth.token_for_code(code)
                hub_user = await self.hub_auth.user_for_token(token, use_cache=False, sync=False)
                async with self.db_lock:
                    db_session.add(
                        UserDB(
                            token=token,
                            name=hub_user["name"],
                            username=hub_user["name"],
                        ),
                    )
                    await db_session.commit()

                next_url = self.hub_auth.get_next_url(cookie_state)
                response = RedirectResponse(next_url)
                response.set_cookie(key="jupyverse_jupyterhub_token", value=token)
                return response

            @router.get("/api/me")
            async def get_api_me(
                request: Request,
                user: User = Depends(self.current_user()),
            ):
                checked_permissions: Dict[str, List[str]] = {}
                permissions = json.loads(
                    dict(request.query_params).get("permissions", "{}").replace("'", '"')
                )
                if permissions:
                    user_permissions: Dict[str, List[str]] = {}
                    for resource, actions in permissions.items():
                        user_resource_permissions = user_permissions.get(resource, [])
                        allowed = checked_permissions[resource] = []
                        for action in actions:
                            if action in user_resource_permissions:
                                allowed.append(action)

                keys = ["username", "name", "display_name", "initials", "avatar_url", "color"]
                identity = {k: getattr(user, k) for k in keys}
                return {
                    "identity": identity,
                    "permissions": checked_permissions,
                }

            self.include_router(router)

        def current_user(self, permissions: Optional[Dict[str, List[str]]] = None) -> Callable:
            async def _(
                request: Request,
                jupyverse_jupyterhub_token: Annotated[Union[str, None], Cookie()] = None,
            ):
                if jupyverse_jupyterhub_token is not None:
                    hub_user = await self.hub_auth.user_for_token(
                        jupyverse_jupyterhub_token, use_cache=False, sync=False
                    )
                    scopes = self.hub_auth.check_scopes(self.hub_auth.access_scopes, hub_user)
                    if not scopes:
                        raise HTTPException(
                            status_code=status.HTTP_403_FORBIDDEN,
                            detail=f"User {hub_user['name']} cannot access this server",
                        )

                    async with self.db_lock:
                        user_db = await db_session.scalar(
                            select(UserDB).filter_by(token=jupyverse_jupyterhub_token)
                        )
                    user = JupyterHubUser.model_validate(user_db)
                    if self.activity_url:
                        headers = {
                            "Authorization": f"token {self.hub_auth.api_token}",
                            "Content-Type": "application/json",
                        }
                        last_activity = isoformat(datetime.utcnow())
                        task = asyncio.create_task(
                            http_client.post(
                                self.activity_url,
                                headers=headers,
                                json={
                                    "servers": {self.server_name: {"last_activity": last_activity}}
                                },
                            )
                        )
                        self.background_tasks.add(task)
                        task.add_done_callback(self.background_tasks.discard)
                    return user

                if permissions:
                    raise HTTPException(
                        status_code=status.HTTP_403_FORBIDDEN,
                    )

                state = self.hub_auth.generate_state(next_url=str(request.url))
                raise HTTPException(
                    status_code=status.HTTP_307_TEMPORARY_REDIRECT,
                    headers={
                        "Location": f"{self.hub_auth.login_url}&state={state}",
                        "Set-Cookie": f"{self.hub_auth.state_cookie_name}={state}",
                    },
                )

            return _

        async def update_user(
            self, jupyverse_jupyterhub_token: Annotated[Union[str, None], Cookie()] = None
        ) -> Callable:
            async def _(data: Dict[str, Any]) -> JupyterHubUser:
                if jupyverse_jupyterhub_token is not None:
                    async with self.db_lock:
                        user_db = await db_session.scalar(
                            select(UserDB).filter_by(token=jupyverse_jupyterhub_token)
                        )
                        for k, v in data.items():
                            setattr(user_db, k, v)
                        await db_session.commit()
                    user = JupyterHubUser.model_validate(user_db)
                    return user

            return _

        def websocket_auth(
            self,
            permissions: Optional[Dict[str, List[str]]] = None,
        ) -> Callable[[], Tuple[Any, Dict[str, List[str]]]]:
            async def _(
                websocket: WebSocket,
            ) -> Optional[Tuple[WebSocket, Optional[Dict[str, List[str]]]]]:
                accept_websocket = False
                if "jupyverse_jupyterhub_token" in websocket._cookies:
                    jupyverse_jupyterhub_token = websocket._cookies["jupyverse_jupyterhub_token"]
                    async with self.db_lock:
                        user_db = await db_session.scalar(
                            select(UserDB).filter_by(token=jupyverse_jupyterhub_token)
                        )
                    if user_db:
                        accept_websocket = True
                if accept_websocket:
                    return websocket, permissions
                else:
                    await websocket.close(code=status.WS_1008_POLICY_VIOLATION)
                    return None

            return _

    return AuthJupyterHub()



import asyncio
import json
import os
import signal
import uuid
from datetime import datetime
from typing import Dict, Iterable, List, Optional, cast

from fastapi import WebSocket, WebSocketDisconnect
from starlette.websockets import WebSocketState

from ..kernel_driver.connect import cfg_t, connect_channel, read_connection_file
from ..kernel_driver.connect import launch_kernel as _launch_kernel
from ..kernel_driver.connect import (
    write_connection_file as _write_connection_file,
)
from ..kernel_driver.message import create_message, receive_message, send_message
from .message import (
    deserialize_msg_from_ws_v1,
    from_binary,
    get_msg_from_parts,
    get_parent_header,
    get_zmq_parts,
    send_raw_message,
    serialize_msg_to_ws_v1,
    to_binary,
)

kernels: dict = {}


class AcceptedWebSocket:
    _websocket: WebSocket
    _accepted_subprotocol: Optional[str]

    def __init__(self, websocket, accepted_subprotocol):
        self._websocket = websocket
        self._accepted_subprotocol = accepted_subprotocol

    @property
    def websocket(self):
        return self._websocket

    @property
    def accepted_subprotocol(self):
        return self._accepted_subprotocol


class KernelServer:
    def __init__(
        self,
        kernelspec_path: str = "",
        kernel_cwd: str = "",
        connection_cfg: Optional[cfg_t] = None,
        connection_file: str = "",
        write_connection_file: bool = True,
        capture_kernel_output: bool = True,
    ) -> None:
        self.capture_kernel_output = capture_kernel_output
        self.kernelspec_path = kernelspec_path
        self.kernel_cwd = kernel_cwd
        self.connection_cfg = connection_cfg
        self.connection_file = connection_file
        self.write_connection_file = write_connection_file
        self.channel_tasks: List[asyncio.Task] = []
        self.sessions: Dict[str, AcceptedWebSocket] = {}
        # blocked messages and allowed messages are mutually exclusive
        self.blocked_messages: List[str] = []
        self.allowed_messages: Optional[List[str]] = None  # when None, all messages are allowed
        # when [], no message is allowed
        self.setup_connection_file()

    def setup_connection_file(self):
        if self.write_connection_file:
            self.connection_file_path, self.connection_cfg = _write_connection_file(
                self.connection_file
            )
        elif self.connection_file:
            self.connection_file_path = self.connection_file
            self.connection_cfg = read_connection_file(self.connection_file)
        else:
            if self.connection_cfg is None:
                raise RuntimeError("No connection_cfg")
        self.key = cast(str, self.connection_cfg["key"])

    def block_messages(self, message_types: Iterable[str] = []):
        # if using blocked messages, discard allowed messages
        self.allowed_messages = None
        if isinstance(message_types, str):
            message_types = [message_types]
        self.blocked_messages = list(message_types)

    def allow_messages(self, message_types: Optional[Iterable[str]] = None):
        # if using allowed messages, discard blocked messages
        self.blocked_messages = []
        if message_types is None:
            self.allowed_messages = None
            return
        if isinstance(message_types, str):
            message_types = [message_types]
        self.allowed_messages = list(message_types)

    @property
    def connections(self) -> int:
        return len(self.sessions)

    async def start(self, launch_kernel: bool = True) -> None:
        self.last_activity = {
            "date": datetime.utcnow().isoformat() + "Z",
            "execution_state": "starting",
        }
        if launch_kernel:
            if not self.kernelspec_path:
                raise RuntimeError("Could not find a kernel, maybe you forgot to install one?")
            self.kernel_process = await _launch_kernel(
                self.kernelspec_path,
                self.connection_file_path,
                self.kernel_cwd,
                self.capture_kernel_output,
            )
        assert self.connection_cfg is not None
        identity = uuid.uuid4().hex.encode("ascii")
        self.shell_channel = connect_channel("shell", self.connection_cfg, identity=identity)
        self.stdin_channel = connect_channel("stdin", self.connection_cfg, identity=identity)
        self.control_channel = connect_channel("control", self.connection_cfg, identity=identity)
        self.iopub_channel = connect_channel("iopub", self.connection_cfg)
        await self._wait_for_ready()
        self.channel_tasks += [
            asyncio.create_task(self.listen("shell")),
            asyncio.create_task(self.listen("stdin")),
            asyncio.create_task(self.listen("control")),
            asyncio.create_task(self.listen("iopub")),
        ]

    async def stop(self) -> None:
        if self.write_connection_file:
            # FIXME: stop kernel in a better way
            try:
                self.kernel_process.send_signal(signal.SIGINT)
                self.kernel_process.kill()
                await self.kernel_process.wait()
            except BaseException:
                pass
            try:
                os.remove(self.connection_file_path)
            except BaseException:
                pass
        for task in self.channel_tasks:
            task.cancel()
        self.channel_tasks = []

    def interrupt(self) -> None:
        self.kernel_process.send_signal(signal.SIGINT)

    async def restart(self) -> None:
        await self.stop()
        self.setup_connection_file()
        await self.start()

    async def serve(
        self,
        websocket: AcceptedWebSocket,
        session_id: str,
        permissions: Optional[Dict[str, List[str]]],
    ):
        self.sessions[session_id] = websocket
        self.can_execute = permissions is None or "execute" in permissions.get("kernels", [])
        await self.listen_web(websocket)
        # the session could have been removed through the REST API, so check if it still exists
        if session_id in self.sessions:
            del self.sessions[session_id]

    async def listen_web(self, websocket: AcceptedWebSocket):
        try:
            await self.send_to_zmq(websocket)
        except WebSocketDisconnect:
            pass

    async def listen(self, channel_name: str):
        if channel_name == "shell":
            channel = self.shell_channel
        elif channel_name == "control":
            channel = self.control_channel
        elif channel_name == "iopub":
            channel = self.iopub_channel
        elif channel_name == "stdin":
            channel = self.stdin_channel

        while True:
            parts = await get_zmq_parts(channel)
            parent_header = get_parent_header(parts)
            if channel == self.iopub_channel:
                # broadcast to all web clients
                for websocket in self.sessions.values():
                    await self.send_to_ws(websocket, parts, parent_header, channel_name)
            else:
                session = parent_header["session"]
                if session in self.sessions:
                    websocket = self.sessions[session]
                    await self.send_to_ws(websocket, parts, parent_header, channel_name)

    async def _wait_for_ready(self):
        while True:
            msg = create_message("kernel_info_request")
            await send_message(msg, self.shell_channel, self.key)
            msg = await receive_message(self.shell_channel, timeout=0.2)
            if msg is not None and msg["msg_type"] == "kernel_info_reply":
                msg = await receive_message(self.iopub_channel, timeout=0.2)
                if msg is None:
                    # IOPub not connected, start over
                    pass
                else:
                    break

    async def send_to_zmq(self, websocket):
        if not websocket.accepted_subprotocol:
            while True:
                msg = await receive_json_or_bytes(websocket.websocket)
                if not self.can_execute:
                    continue
                msg_type = msg["header"]["msg_type"]
                if (msg_type in self.blocked_messages) or (
                    self.allowed_messages is not None and msg_type not in self.allowed_messages
                ):
                    continue
                channel = msg.pop("channel")
                if channel == "shell":
                    await send_message(msg, self.shell_channel, self.key)
                elif channel == "control":
                    await send_message(msg, self.control_channel, self.key)
                elif channel == "stdin":
                    await send_message(msg, self.stdin_channel, self.key)
        elif websocket.accepted_subprotocol == "v1.kernel.websocket.jupyter.org":
            while True:
                msg = await websocket.websocket.receive_bytes()
                if not self.can_execute:
                    continue
                channel, parts = deserialize_msg_from_ws_v1(msg)
                # NOTE: we parse the header for message filtering
                # it is not as bad as parsing the content
                header = json.loads(parts[0])
                msg_type = header["msg_type"]
                if (msg_type in self.blocked_messages) or (
                    self.allowed_messages is not None and msg_type not in self.allowed_messages
                ):
                    continue
                if channel == "shell":
                    await send_raw_message(parts, self.shell_channel, self.key)
                elif channel == "control":
                    await send_raw_message(parts, self.control_channel, self.key)
                elif channel == "stdin":
                    await send_raw_message(parts, self.stdin_channel, self.key)

    async def send_to_ws(self, websocket, parts, parent_header, channel_name):
        if not websocket.accepted_subprotocol:
            # default, "legacy" protocol
            msg = get_msg_from_parts(parts, parent_header=parent_header)
            msg["channel"] = channel_name
            await send_json_or_bytes(websocket.websocket, msg)
            if channel_name == "iopub":
                if "content" in msg and "execution_state" in msg["content"]:
                    self.last_activity = {
                        "date": msg["header"]["date"],
                        "execution_state": msg["content"]["execution_state"],
                    }
        elif websocket.accepted_subprotocol == "v1.kernel.websocket.jupyter.org":
            bin_msg = serialize_msg_to_ws_v1(parts, channel_name)
            try:
                await websocket.websocket.send_bytes(bin_msg)
            except BaseException:
                pass
            # FIXME: update last_activity
            # but we don't want to parse the content!
            # or should we request it from the control channel?


async def receive_json_or_bytes(websocket):
    assert websocket.application_state == WebSocketState.CONNECTED
    message = await websocket.receive()
    websocket._raise_on_disconnect(message)
    if "text" in message:
        return json.loads(message["text"])
    msg = from_binary(message["bytes"])
    return msg


async def send_json_or_bytes(websocket, msg):
    bmsg = to_binary(msg)
    if bmsg is None:
        await websocket.send_json(msg)
    else:
        await websocket.send_bytes(bmsg)



import os
from urllib.parse import unquote, urlparse

from jupyverse_api.cli import main


def launch():
    service_url = unquote(os.environ.get("JUPYTERHUB_SERVICE_URL"))
    url = urlparse(service_url)
    return main.callback(
        open_browser=True,
        host=url.hostname,
        port=url.port,
        set_=(
            f"frontend.base_url={url.path}",
            f"app.mount_path={url.path}",
        ),
        disable=[],
    )



from typing import cast

from fastapi import Request
from fastapi.responses import RedirectResponse


class RedirectException(Exception):
    def __init__(self, redirect_to: str):
        self.redirect_to = redirect_to


async def _redirect_exception_handler(request: Request, exc: Exception) -> RedirectResponse:
    exc = cast(RedirectException, exc)
    return RedirectResponse(url=exc.redirect_to)



from __future__ import annotations

from inspect import isawaitable
from typing import Any, Awaitable, Callable

from .websocket_server import WebsocketServer


class ASGIWebsocket:
    def __init__(
        self,
        receive: Callable[[], Awaitable[dict[str, Any]]],
        send: Callable[[dict[str, Any]], Awaitable[None]],
        path: str,
        on_disconnect: Callable[[dict[str, Any]], Awaitable[None] | None] | None = None,
    ):
        self._receive = receive
        self._send = send
        self._path = path
        self._on_disconnect = on_disconnect

    @property
    def path(self) -> str:
        return self._path

    def __aiter__(self):
        return self

    async def __anext__(self) -> bytes:
        return await self.recv()

    async def send(self, message: bytes) -> None:
        await self._send(
            dict(
                type="websocket.send",
                bytes=message,
            )
        )

    async def recv(self) -> bytes:
        message = await self._receive()
        if message["type"] == "websocket.receive":
            return message["bytes"]
        if message["type"] == "websocket.disconnect":
            if self._on_disconnect is not None:
                res = self._on_disconnect(message)
                if isawaitable(res):
                    await res
            raise StopAsyncIteration()
        return b""


class ASGIServer:
    """ASGI server."""

    def __init__(
        self,
        websocket_server: WebsocketServer,
        on_connect: Callable[[dict[str, Any], dict[str, Any]], Awaitable[bool] | bool]
        | None = None,
        on_disconnect: Callable[[dict[str, Any]], Awaitable[None] | None] | None = None,
    ):
        """Initialize the object.

        Arguments:
            websocket_server: An instance of WebsocketServer.
            on_connect: An optional callback to call when connecting the WebSocket.
                If the callback returns True, the WebSocket is not accepted.
            on_disconnect: An optional callback called when disconnecting the WebSocket.
        """
        self._websocket_server = websocket_server
        self._on_connect = on_connect
        self._on_disconnect = on_disconnect

    async def __call__(
        self,
        scope: dict[str, Any],
        receive: Callable[[], Awaitable[dict[str, Any]]],
        send: Callable[[dict[str, Any]], Awaitable[None]],
    ):
        msg = await receive()
        if msg["type"] == "websocket.connect":
            if self._on_connect is not None:
                close = self._on_connect(msg, scope)
                if isawaitable(close):
                    close = await close
                if close:
                    return

            await send({"type": "websocket.accept"})
            websocket = ASGIWebsocket(receive, send, scope["path"], self._on_disconnect)
            await self._websocket_server.serve(websocket)



from typing import List, Tuple

import pkg_resources
import rich_click as click
from asphalt.core.cli import run


@click.command()  # type: ignore
@click.option(
    "--open-browser",
    is_flag=True,
    show_default=True,
    default=False,
    help="Open a browser window.",
)
@click.option(
    "--host",
    type=str,
    default="127.0.0.1",
    help="The host URL.",
)
@click.option(
    "--port",
    type=int,
    default=8000,
    help="The host port.",
)
@click.option(
    "--allow-origin",
    multiple=True,
    type=str,
    help="The origin to allow.",
)
@click.option(
    "--set",
    "set_",
    multiple=True,
    type=str,
    help="Set configuration.",
)
@click.option(
    "--disable",
    multiple=True,
    type=str,
    help="Disable plugin.",
)
def main(
    open_browser: bool = False,
    host: str = "127.0.0.1",
    port: int = 8000,
    set_: Tuple[str, ...] = (),
    disable: Tuple[str, ...] = (),
    allow_origin: Tuple[str, ...] = (),
) -> None:
    set_list: List[str] = list(set_)
    for i, s in enumerate(set_list):
        set_list[i] = f"component.components.{s}"
    set_list.append(f"component.open_browser={open_browser}")
    set_list.append(f"component.host={host}")
    set_list.append(f"component.port={port}")
    set_list.append(f"component.allow_origin={allow_origin}")
    config = get_config(disable)
    run.callback(
        unsafe=False,
        loop=None,
        set_=set_list,
        service=None,
        configfile=[config],
    )  # type: ignore


def get_config(disable: Tuple[str, ...]) -> str:
    jupyverse_components = [
        ep.name
        for ep in pkg_resources.iter_entry_points(group="jupyverse.components")
        if ep.name not in disable
    ]

    config = ["component:\n  type: jupyverse\n  components:\n"]
    for component in jupyverse_components:
        config.append(f"    {component}:\n      type: {component}\n")

    config.append(
        """
logging:
  version: 1
  disable_existing_loggers: false
  formatters:
    default:
      format: '[%(asctime)s %(levelname)s] %(message)s'
  handlers:
    console:
      class: logging.StreamHandler
      formatter: default
  root:
    handlers: [console]
    level: INFO
  loggers:
    webnotifier:
      level: DEBUG
        """
    )

    config_str = "".join(config)
    return config_str



import json

import pytest
from asphalt.core import Context
from httpx import AsyncClient
from utils import configure

from jupyverse_api.main import JupyverseComponent

test_theme = {"raw": '{// jupyverse test\n"theme": "JupyterLab Dark"}'}

COMPONENTS = {
    "app": {"type": "app"},
    "auth": {"type": "auth", "test": True},
    "contents": {"type": "contents"},
    "frontend": {"type": "frontend"},
    "lab": {"type": "lab"},
    "jupyterlab": {"type": "jupyterlab"},
    "kernels": {"type": "kernels"},
    "yjs": {"type": "yjs"},
}


@pytest.mark.asyncio
@pytest.mark.parametrize("auth_mode", ("noauth",))
async def test_settings(auth_mode, unused_tcp_port):
    components = configure(COMPONENTS, {"auth": {"mode": auth_mode}})
    async with Context() as ctx, AsyncClient() as http:
        await JupyverseComponent(
            components=components,
            port=unused_tcp_port,
        ).start(ctx)

        # get previous theme
        response = await http.get(
            f"http://127.0.0.1:{unused_tcp_port}/lab/api/settings/@jupyterlab/apputils-extension:themes"
        )
        assert response.status_code == 200
        theme = {"raw": json.loads(response.content)["raw"]}
        # put new theme
        response = await http.put(
            f"http://127.0.0.1:{unused_tcp_port}/lab/api/settings/@jupyterlab/apputils-extension:themes",
            data=json.dumps(test_theme),
        )
        assert response.status_code == 204
        # get new theme
        response = await http.get(
            f"http://127.0.0.1:{unused_tcp_port}/lab/api/settings/@jupyterlab/apputils-extension:themes"
        )
        assert response.status_code == 200
        assert json.loads(response.content)["raw"] == test_theme["raw"]
        # put previous theme back
        response = await http.put(
            f"http://127.0.0.1:{unused_tcp_port}/lab/api/settings/@jupyterlab/apputils-extension:themes",
            data=json.dumps(theme),
        )
        assert response.status_code == 204



import json
from pathlib import Path

import retrolab  # type: ignore
from fastapi.staticfiles import StaticFiles

from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User
from jupyverse_api.frontend import FrontendConfig
from jupyverse_api.lab import Lab
from jupyverse_api.retrolab import RetroLab


class _RetroLab(RetroLab):
    def __init__(self, app: App, auth: Auth, frontend_config: FrontendConfig, lab: Lab) -> None:
        super().__init__(app, auth, lab)
        self.frontend_config = frontend_config
        self.lab = lab
        self.lab.redirect_after_root = "retro/tree"

        self.retrolab_dir = Path(retrolab.__file__).parent

        self.mount(
            "/static/retro",
            StaticFiles(directory=self.retrolab_dir / "static"),
            name="static",
        )

        for path in (self.retrolab_dir / "labextension" / "static").glob("remoteEntry.*.js"):
            load = f"static/{path.name}"
            break

        self.retro_federated_extensions = [
            {
                "extension": "./extension",
                "load": load,
                "name": "@retrolab/lab-extension",
                "style": "./style",
            }
        ]

    async def get_tree(
        self,
        user: User,
    ):
        return get_index(
            self.lab,
            self.retrolab_dir,
            self.retro_federated_extensions,
            "Tree",
            "tree",
            self.frontend_config.collaborative,
            self.frontend_config.base_url,
        )

    async def get_notebook(
        self,
        path,
        user: User,
    ):
        return get_index(
            self.lab,
            self.retrolab_dir,
            self.retro_federated_extensions,
            path,
            "notebooks",
            self.frontend_config.collaborative,
            self.frontend_config.base_url,
        )

    async def edit_file(
        self,
        path,
        user: User,
    ):
        return get_index(
            self.lab,
            self.retrolab_dir,
            self.retro_federated_extensions,
            path,
            "edit",
            self.frontend_config.collaborative,
            self.frontend_config.base_url,
        )

    async def get_console(
        self,
        path,
        user: User,
    ):
        return get_index(
            self.lab,
            self.retrolab_dir,
            self.retro_federated_extensions,
            path,
            "consoles",
            self.frontend_config.collaborative,
            self.frontend_config.base_url,
        )

    async def get_terminal(
        self,
        name: str,
        user: User,
    ):
        return get_index(
            self.lab,
            self.retrolab_dir,
            self.retro_federated_extensions,
            name,
            "terminals",
            self.frontend_config.collaborative,
            self.frontend_config.base_url,
        )


INDEX_HTML = """\
<!DOCTYPE html>
<html>
<head>
  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <title>RetroLab - DOC_NAME</title>
  <link rel="icon" type="image/x-icon" href="/static/favicons/favicon-notebook.ico"
    class="favicon">
</head>
<body>
  <script id="jupyter-config-data" type="application/json">
    PAGE_CONFIG
  </script>
  <script src="BASE_URLstatic/retro/bundle.js" main="index"></script>
  <script type="text/javascript">
    /* Remove token from URL. */
    (function () {
      var parsedUrl = new URL(window.location.href);
      if (parsedUrl.searchParams.get('token')) {
        parsedUrl.searchParams.delete('token');
        window.history.replaceState({ }, '', parsedUrl.href);
      }
    })();
  </script>
</body>
</html>
"""


def get_index(
    lab, retrolab_dir, retro_federated_extensions, doc_name, retro_page, collaborative, base_url="/"
):
    extensions_dir = lab.prefix_dir / "share" / "jupyter" / "labextensions"
    federated_extensions, disabled_extension = lab.get_federated_extensions(extensions_dir)
    page_config = {
        "appName": "RetroLab",
        "appNamespace": "retro",
        "appSettingsDir": (lab.prefix_dir / "share" / "jupyter" / "lab" / "settings").as_posix(),
        "appUrl": "/lab",
        "appVersion": retrolab.__version__,
        "baseUrl": base_url,
        "cacheFiles": True,
        "collaborative": collaborative,
        "disabledExtensions": disabled_extension,
        "extraLabextensionsPath": [],
        "federated_extensions": retro_federated_extensions + federated_extensions,
        "frontendUrl": "/retro/",
        "fullAppUrl": f"{base_url}lab",
        "fullLabextensionsUrl": f"{base_url}lab/extensions",
        "fullLicensesUrl": f"{base_url}lab/api/licenses",
        "fullListingsUrl": f"{base_url}lab/api/listings",
        "fullMathjaxUrl": f"{base_url}static/notebook/components/MathJax/MathJax.js",
        "fullSettingsUrl": f"{base_url}lab/api/settings",
        "fullStaticUrl": f"{base_url}static/retro",
        "fullThemesUrl": f"{base_url}lab/api/themes",
        "fullTranslationsApiUrl": f"{base_url}lab/api/translations",
        "fullTreeUrl": f"{base_url}lab/tree",
        "fullWorkspacesApiUrl": f"{base_url}lab/api/workspaces",
        "labextensionsPath": [(lab.prefix_dir / "share" / "jupyter" / "labextensions").as_posix()],
        "labextensionsUrl": "/lab/extensions",
        "licensesUrl": "/lab/api/licenses",
        "listingsUrl": "/lab/api/listings",
        "mathjaxConfig": "TeX-AMS-MML_HTMLorMML-full,Safe",
        "retroLogo": False,
        "retroPage": retro_page,
        "schemasDir": (lab.prefix_dir / "share" / "jupyter" / "lab" / "schemas").as_posix(),
        "settingsUrl": "/lab/api/settings",
        "staticDir": (retrolab_dir / "static").as_posix(),
        "templatesDir": (retrolab_dir / "templates").as_posix(),
        "terminalsAvailable": True,
        "themesDir": (lab.prefix_dir / "share" / "jupyter" / "lab" / "themes").as_posix(),
        "themesUrl": "/lab/api/themes",
        "translationsApiUrl": "/lab/api/translations",
        "treeUrl": "/lab/tree",
        "workspacesApiUrl": "/lab/api/workspaces",
        "wsUrl": "",
    }
    index = (
        INDEX_HTML.replace("PAGE_CONFIG", json.dumps(page_config))
        .replace("DOC_NAME", doc_name)
        .replace("BASE_URL", base_url)
    )
    return index



import pytest
from asphalt.core import Context
from httpx import AsyncClient
from httpx_ws import WebSocketUpgradeError, aconnect_ws
from utils import authenticate_client, configure

from jupyverse_api.auth import AuthConfig
from jupyverse_api.main import JupyverseComponent

COMPONENTS = {
    "app": {"type": "app"},
    "auth": {"type": "auth", "test": True},
    "contents": {"type": "contents"},
    "frontend": {"type": "frontend"},
    "lab": {"type": "lab"},
    "jupyterlab": {"type": "jupyterlab"},
    "kernels": {"type": "kernels"},
    "yjs": {"type": "yjs"},
}


@pytest.mark.asyncio
async def test_kernel_channels_unauthenticated(unused_tcp_port):
    async with Context() as ctx:
        await JupyverseComponent(
            components=COMPONENTS,
            port=unused_tcp_port,
        ).start(ctx)

        with pytest.raises(WebSocketUpgradeError):
            async with aconnect_ws(
                f"http://127.0.0.1:{unused_tcp_port}/api/kernels/kernel_id_0/channels?session_id=session_id_0",
            ):
                pass


@pytest.mark.asyncio
async def test_kernel_channels_authenticated(unused_tcp_port):
    async with Context() as ctx, AsyncClient() as http:
        await JupyverseComponent(
            components=COMPONENTS,
            port=unused_tcp_port,
        ).start(ctx)

        await authenticate_client(http, unused_tcp_port)
        async with aconnect_ws(
            f"http://127.0.0.1:{unused_tcp_port}/api/kernels/kernel_id_0/channels?session_id=session_id_0",
            http,
        ):
            pass


@pytest.mark.asyncio
@pytest.mark.parametrize("auth_mode", ("noauth", "token", "user"))
async def test_root_auth(auth_mode, unused_tcp_port):
    components = configure(COMPONENTS, {"auth": {"mode": auth_mode}})
    async with Context() as ctx, AsyncClient() as http:
        await JupyverseComponent(
            components=components,
            port=unused_tcp_port,
        ).start(ctx)

        response = await http.get(f"http://127.0.0.1:{unused_tcp_port}/")
        if auth_mode == "noauth":
            expected = 302
        elif auth_mode in ["token", "user"]:
            expected = 403

        assert response.status_code == expected
        assert response.headers["content-type"] == "application/json"


@pytest.mark.asyncio
@pytest.mark.parametrize("auth_mode", ("noauth",))
async def test_no_auth(auth_mode, unused_tcp_port):
    components = configure(COMPONENTS, {"auth": {"mode": auth_mode}})
    async with Context() as ctx, AsyncClient() as http:
        await JupyverseComponent(
            components=components,
            port=unused_tcp_port,
        ).start(ctx)

        response = await http.get(f"http://127.0.0.1:{unused_tcp_port}/lab")
        assert response.status_code == 200


@pytest.mark.asyncio
@pytest.mark.parametrize("auth_mode", ("token",))
async def test_token_auth(auth_mode, unused_tcp_port):
    components = configure(COMPONENTS, {"auth": {"mode": auth_mode}})
    async with Context() as ctx, AsyncClient() as http:
        await JupyverseComponent(
            components=components,
            port=unused_tcp_port,
        ).start(ctx)

        auth_config = await ctx.request_resource(AuthConfig)

        # no token provided, should not work
        response = await http.get(f"http://127.0.0.1:{unused_tcp_port}/")
        assert response.status_code == 403
        # token provided, should work
        response = await http.get(f"http://127.0.0.1:{unused_tcp_port}/?token={auth_config.token}")
        assert response.status_code == 302


@pytest.mark.asyncio
@pytest.mark.parametrize("auth_mode", ("user",))
@pytest.mark.parametrize(
    "permissions",
    (
        {},
        {"admin": ["read"], "foo": ["bar", "baz"]},
    ),
)
async def test_permissions(auth_mode, permissions, unused_tcp_port):
    components = configure(COMPONENTS, {"auth": {"mode": auth_mode}})
    async with Context() as ctx, AsyncClient() as http:
        await JupyverseComponent(
            components=components,
            port=unused_tcp_port,
        ).start(ctx)

        await authenticate_client(http, unused_tcp_port, permissions=permissions)
        response = await http.get(f"http://127.0.0.1:{unused_tcp_port}/auth/user/me")
        if "admin" in permissions.keys():
            # we have the permissions
            assert response.status_code == 200
        else:
            # we don't have the permissions
            assert response.status_code == 403



from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Optional

from pycrdt import Doc, Map


class YBaseDoc(ABC):
    def __init__(self, ydoc: Optional[Doc] = None):
        if ydoc is None:
            self._ydoc = Doc()
        else:
            self._ydoc = ydoc
        self._ystate = Map()
        self._ydoc["state"] = self._ystate
        self._subscriptions: Dict[Any, str] = {}

    @property
    @abstractmethod
    def version(self) -> str:
        ...

    @property
    def ystate(self) -> Map:
        return self._ystate

    @property
    def ydoc(self) -> Doc:
        return self._ydoc

    @property
    def source(self) -> Any:
        return self.get()

    @source.setter
    def source(self, value: Any):
        return self.set(value)

    @property
    def dirty(self) -> Optional[bool]:
        return self._ystate.get("dirty")

    @dirty.setter
    def dirty(self, value: bool) -> None:
        self._ystate["dirty"] = value

    @property
    def path(self) -> Optional[str]:
        return self._ystate.get("path")

    @path.setter
    def path(self, value: str) -> None:
        self._ystate["path"] = value

    @property
    def file_id(self) -> Optional[str]:
        return self._ystate.get("file_id")

    @file_id.setter
    def file_id(self, value: str) -> None:
        self._ystate["file_id"] = value

    @abstractmethod
    def get(self) -> Any:
        ...

    @abstractmethod
    def set(self, value: Any) -> None:
        ...

    @abstractmethod
    def observe(self, callback: Callable[[str, Any], None]) -> None:
        ...

    def unobserve(self) -> None:
        for k, v in self._subscriptions.items():
            k.unobserve(v)
        self._subscriptions = {}



import sys
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Optional, Tuple

from fastapi import APIRouter, Depends, Request, Response
from fastapi.staticfiles import StaticFiles

from jupyverse_api import Router
from jupyverse_api.jupyterlab import JupyterLabConfig

from ..app import App
from ..auth import Auth, User


class Lab(Router, ABC):
    prefix_dir: Path
    jlab_dir: Path
    labextensions_dir: Path
    extensions_dir: Path
    redirect_after_root: str

    def __init__(self, app: App, auth: Auth, jupyterlab_config: Optional[JupyterLabConfig]):
        super().__init__(app)

        self.prefix_dir = Path(sys.prefix)
        self.extensions_dir = self.prefix_dir / "share" / "jupyter" / "labextensions"
        self.federated_extensions, disabled_extensions = self.get_federated_extensions(
            self.extensions_dir
        )
        if jupyterlab_config is not None:
            import jupyterlab as jupyterlab_module  # type: ignore

            jlab_dev_mode = jupyterlab_config.dev_mode
        else:
            jlab_dev_mode = False
        if jlab_dev_mode:
            self.jlab_dir = Path(jupyterlab_module.__file__).parents[1] / "dev_mode"
        else:
            self.jlab_dir = self.prefix_dir / "share" / "jupyter" / "lab"
        self.labextensions_dir = self.prefix_dir / "share" / "jupyter" / "labextensions"
        for ext in self.federated_extensions:
            name = ext["name"]
            self.mount(
                f"/lab/extensions/{name}/static",
                StaticFiles(directory=self.extensions_dir / name / "static"),
                name=name,
            )

        self.mount(
            "/lab/api/themes",
            StaticFiles(directory=self.jlab_dir / "themes"),
            name="themes",
        )

        router = APIRouter()

        @router.get("/", name="root")
        async def get_root(
            response: Response,
            user: User = Depends(auth.current_user()),
        ):
            return await self.get_root(response, user)

        @router.get("/favicon.ico")
        async def get_favicon():
            return await self.get_favicon()

        @router.get("/static/notebook/components/MathJax/{rest_of_path:path}")
        async def get_mathjax(rest_of_path):
            return await self.get_mathjax(rest_of_path)

        @router.get("/lab/api/listings/@jupyterlab/extensionmanager-extension/listings.json")
        async def get_listings(user: User = Depends(auth.current_user())):
            return await self.get_listings(user)

        @router.get("/lab/api/extensions")
        async def get_extensions(user: User = Depends(auth.current_user())):
            return await self.get_extensions(user)

        @router.get("/lab/api/translations/")
        async def get_translations_(
            user: User = Depends(auth.current_user()),
        ):
            return await self.get_translations_(user)

        @router.get("/lab/api/translations")
        async def get_translations(user: User = Depends(auth.current_user())):
            return await self.get_translations(user)

        @router.get("/lab/api/translations/{language}")
        async def get_translation(
            language,
            user: User = Depends(auth.current_user()),
        ):
            return await self.get_translation(language, user)

        @router.get("/lab/api/settings/{name0}/{name1}:{name2}")
        async def get_setting(
            name0,
            name1,
            name2,
            user: User = Depends(auth.current_user()),
        ):
            return await self.get_setting(name0, name1, name2, user)

        @router.put(
            "/lab/api/settings/@jupyterlab/{name0}:{name1}",
            status_code=204,
        )
        async def change_setting(
            request: Request,
            name0,
            name1,
            user: User = Depends(auth.current_user()),
            user_update=Depends(auth.update_user),
        ):
            return await self.change_setting(request, name0, name1, user, user_update)

        @router.get("/lab/api/settings")
        async def get_settings(user: User = Depends(auth.current_user())):
            return await self.get_settings(user)

        self.include_router(router)

    @abstractmethod
    def get_federated_extensions(self, extensions_dir: Path) -> Tuple[List, List]:
        ...

    @abstractmethod
    async def get_root(
        self,
        response: Response,
        user: User,
    ):
        ...

    @abstractmethod
    async def get_favicon(self):
        ...

    @abstractmethod
    async def get_mathjax(self, rest_of_path):
        ...

    @abstractmethod
    async def get_listings(self, user: User):
        ...

    @abstractmethod
    async def get_extensions(self, user: User):
        ...

    @abstractmethod
    async def get_translations_(
        self,
        user: User,
    ):
        ...

    @abstractmethod
    async def get_translations(self, user: User):
        ...

    @abstractmethod
    async def get_translation(
        self,
        language,
        user: User,
    ):
        ...

    @abstractmethod
    async def get_setting(
        self,
        name0,
        name1,
        name2,
        user: User,
    ):
        ...

    @abstractmethod
    async def change_setting(
        self,
        request: Request,
        name0,
        name1,
        user: User,
        user_update,
    ):
        ...

    @abstractmethod
    async def get_settings(self, user: User):
        ...



import logging
import secrets
from dataclasses import dataclass
from pathlib import Path
from typing import Any, AsyncGenerator, List

from fastapi import Depends
from fastapi_users.db import (
    SQLAlchemyBaseOAuthAccountTableUUID,
    SQLAlchemyBaseUserTableUUID,
    SQLAlchemyUserDatabase,
)
from sqlalchemy import JSON, Boolean, Column, String, Text
from sqlalchemy.ext.asyncio import AsyncAttrs, AsyncSession, async_sessionmaker, create_async_engine
from sqlalchemy.orm import DeclarativeBase, Mapped, relationship

from .config import _AuthConfig

logger = logging.getLogger("auth")


class Base(AsyncAttrs, DeclarativeBase):
    pass


class OAuthAccount(SQLAlchemyBaseOAuthAccountTableUUID, Base):
    pass


class User(SQLAlchemyBaseUserTableUUID, Base):
    anonymous = Column(Boolean, default=True, nullable=False)
    username = Column(String(length=32), nullable=False, unique=True)
    name = Column(String(length=32), default="")
    display_name = Column(String(length=32), default="")
    initials = Column(String(length=8), nullable=True)
    color = Column(String(length=32), nullable=True)
    avatar_url = Column(String(length=32), nullable=True)
    workspace = Column(Text(), default="{}", nullable=False)
    settings = Column(Text(), default="{}", nullable=False)
    permissions = Column(JSON, default={}, nullable=False)
    oauth_accounts: Mapped[List[OAuthAccount]] = relationship("OAuthAccount", lazy="joined")


@dataclass
class Res:
    User: Any
    async_session_maker: Any
    create_db_and_tables: Any
    get_async_session: Any
    get_user_db: Any
    secret: Any


def get_db(auth_config: _AuthConfig) -> Res:
    jupyter_dir = (
        Path.home() / ".local" / "share" / "jupyter"
        if auth_config.directory is None
        else Path(auth_config.directory)
    )
    jupyter_dir.mkdir(parents=True, exist_ok=True)
    name = "jupyverse"
    if auth_config.test:
        name += "_test"
    secret_path = jupyter_dir / f"{name}_secret"
    userdb_path = jupyter_dir / f"{name}_users.db"

    if auth_config.clear_users:
        if userdb_path.is_file():
            userdb_path.unlink()
        if secret_path.is_file():
            secret_path.unlink()
    if auth_config.mode == "token":
        if secret_path.is_file():
            secret_path.unlink()

    if not secret_path.is_file():
        secret_path.write_text(secrets.token_hex(32))

    secret = secret_path.read_text()

    database_url = f"sqlite+aiosqlite:///{userdb_path}"

    engine = create_async_engine(database_url)
    async_session_maker = async_sessionmaker(engine, expire_on_commit=False)

    async def create_db_and_tables():
        async with engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)

    async def get_async_session() -> AsyncGenerator[AsyncSession, None]:
        async with async_session_maker() as session:
            yield session

    async def get_user_db(session: AsyncSession = Depends(get_async_session)):
        yield SQLAlchemyUserDatabase(session, User, OAuthAccount)

    return Res(
        User=User,
        async_session_maker=async_session_maker,
        create_db_and_tables=create_db_and_tables,
        get_async_session=get_async_session,
        get_user_db=get_user_db,
        secret=secret,
    )



from sqlalchemy import JSON, Boolean, Column, String, Text
from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase


class Base(AsyncAttrs, DeclarativeBase):
    pass


class UserDB(Base):
    __tablename__ = "user_account"

    token = Column(String(32), primary_key=True)
    anonymous = Column(Boolean, default=True, nullable=False)
    username = Column(String(length=32), nullable=False, unique=True)
    name = Column(String(length=32), default="")
    display_name = Column(String(length=32), default="")
    initials = Column(String(length=8), nullable=True)
    color = Column(String(length=32), nullable=True)
    avatar_url = Column(String(length=32), nullable=True)
    workspace = Column(Text(), default="{}", nullable=False)
    settings = Column(Text(), default="{}", nullable=False)
    permissions = Column(JSON, default={}, nullable=False)



from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional, Tuple

from jupyverse_api import Config

from .models import User  # noqa


class Auth(ABC):
    @abstractmethod
    def current_user(self, permissions: Optional[Dict[str, List[str]]] = None) -> Callable:
        ...

    @abstractmethod
    async def update_user(self) -> Callable:
        ...

    @abstractmethod
    def websocket_auth(
        self,
        permissions: Optional[Dict[str, List[str]]] = None,
    ) -> Callable[[], Tuple[Any, Dict[str, List[str]]]]:
        ...


class AuthConfig(Config):
    pass



import copy
import json
from functools import partial
from typing import Any, Callable, Dict, Optional
from uuid import uuid4

from pycrdt import Array, Doc, Map, Text

from .utils import cast_all
from .ybasedoc import YBaseDoc

# The default major version of the notebook format.
NBFORMAT_MAJOR_VERSION = 4
# The default minor version of the notebook format.
NBFORMAT_MINOR_VERSION = 5


class YNotebook(YBaseDoc):
    def __init__(self, ydoc: Optional[Doc] = None):
        super().__init__(ydoc)
        self._ymeta = Map()
        self._ycells = Array()
        self._ydoc["meta"] = self._ymeta
        self._ydoc["cells"] = self._ycells

    @property
    def version(self) -> str:
        return "1.0.0"

    @property
    def ycells(self):
        return self._ycells

    @property
    def cell_number(self) -> int:
        return len(self._ycells)

    def get_cell(self, index: int) -> Dict[str, Any]:
        meta = json.loads(str(self._ymeta))
        cell = json.loads(str(self._ycells[index]))
        cell.pop("execution_status", None)
        cast_all(cell, float, int)  # cells coming from Yjs have e.g. execution_count as float
        if "id" in cell and meta["nbformat"] == 4 and meta["nbformat_minor"] <= 4:
            # strip cell IDs if we have notebook format 4.0-4.4
            del cell["id"]
        if (
            "attachments" in cell
            and cell["cell_type"] in ("raw", "markdown")
            and not cell["attachments"]
        ):
            del cell["attachments"]
        return cell

    def append_cell(self, value: Dict[str, Any]) -> None:
        ycell = self.create_ycell(value)
        self._ycells.append(ycell)

    def set_cell(self, index: int, value: Dict[str, Any]) -> None:
        ycell = self.create_ycell(value)
        self.set_ycell(index, ycell)

    def create_ycell(self, value: Dict[str, Any]) -> Map:
        cell = copy.deepcopy(value)
        if "id" not in cell:
            cell["id"] = str(uuid4())
        cell_type = cell["cell_type"]
        cell_source = cell["source"]
        cell_source = "".join(cell_source) if isinstance(cell_source, list) else cell_source
        cell["source"] = Text(cell_source)
        cell["metadata"] = Map(cell.get("metadata", {}))

        if cell_type in ("raw", "markdown"):
            if "attachments" in cell and not cell["attachments"]:
                del cell["attachments"]
        elif cell_type == "code":
            cell["outputs"] = Array(cell.get("outputs", []))
            cell["execution_status"] = "idle"

        return Map(cell)

    def set_ycell(self, index: int, ycell: Map) -> None:
        self._ycells[index] = ycell

    def get(self) -> Dict:
        meta = json.loads(str(self._ymeta))
        cast_all(meta, float, int)  # notebook coming from Yjs has e.g. nbformat as float
        cells = []
        for i in range(len(self._ycells)):
            cell = self.get_cell(i)
            if "id" in cell and meta["nbformat"] == 4 and meta["nbformat_minor"] <= 4:
                # strip cell IDs if we have notebook format 4.0-4.4
                del cell["id"]
            if (
                "attachments" in cell
                and cell["cell_type"] in ["raw", "markdown"]
                and not cell["attachments"]
            ):
                del cell["attachments"]
            cells.append(cell)

        return dict(
            cells=cells,
            metadata=meta.get("metadata", {}),
            nbformat=int(meta.get("nbformat", 0)),
            nbformat_minor=int(meta.get("nbformat_minor", 0)),
        )

    def set(self, value: Dict) -> None:
        nb_without_cells = {key: value[key] for key in value.keys() if key != "cells"}
        nb = copy.deepcopy(nb_without_cells)
        cast_all(nb, int, float)  # Yjs expects numbers to be floating numbers
        cells = value["cells"] or [
            {
                "cell_type": "code",
                "execution_count": None,
                # auto-created empty code cell without outputs ought be trusted
                "metadata": {"trusted": True},
                "outputs": [],
                "source": "",
                "id": str(uuid4()),
            }
        ]

        with self._ydoc.transaction():
            # clear document
            self._ymeta.clear()
            self._ycells.clear()
            for key in [k for k in self._ystate.keys() if k not in ("dirty", "path", "file_id")]:
                del self._ystate[key]

            # initialize document
            self._ycells.extend([self.create_ycell(cell) for cell in cells])
            self._ymeta["nbformat"] = nb.get("nbformat", NBFORMAT_MAJOR_VERSION)
            self._ymeta["nbformat_minor"] = nb.get("nbformat_minor", NBFORMAT_MINOR_VERSION)

            metadata = nb.get("metadata", {})
            metadata.setdefault("language_info", {"name": ""})
            metadata.setdefault("kernelspec", {"name": "", "display_name": ""})

            self._ymeta["metadata"] = Map(metadata)

    def observe(self, callback: Callable[[str, Any], None]) -> None:
        self.unobserve()
        self._subscriptions[self._ystate] = self._ystate.observe(partial(callback, "state"))
        self._subscriptions[self._ymeta] = self._ymeta.observe_deep(partial(callback, "meta"))
        self._subscriptions[self._ycells] = self._ycells.observe_deep(partial(callback, "cells"))



import os
import sys
from pathlib import Path
from time import sleep

import pytest
from asphalt.core import Context
from fps_kernels.kernel_server.server import KernelServer, kernels
from httpx import AsyncClient
from httpx_ws import aconnect_ws
from utils import configure

from jupyverse_api.main import JupyverseComponent

os.environ["PYDEVD_DISABLE_FILE_VALIDATION"] = "1"

COMPONENTS = {
    "app": {"type": "app"},
    "auth": {"type": "auth", "test": True},
    "contents": {"type": "contents"},
    "frontend": {"type": "frontend"},
    "lab": {"type": "lab"},
    "jupyterlab": {"type": "jupyterlab"},
    "kernels": {"type": "kernels"},
    "yjs": {"type": "yjs"},
}


@pytest.mark.asyncio
@pytest.mark.parametrize("auth_mode", ("noauth",))
async def test_kernel_messages(auth_mode, capfd, unused_tcp_port):
    kernel_id = "kernel_id_0"
    kernel_name = "python3"
    kernelspec_path = (
        Path(sys.prefix) / "share" / "jupyter" / "kernels" / kernel_name / "kernel.json"
    )
    assert kernelspec_path.exists()
    kernel_server = KernelServer(kernelspec_path=kernelspec_path, capture_kernel_output=False)
    await kernel_server.start()
    kernels[kernel_id] = {"server": kernel_server}
    msg_id = "0"
    msg = {
        "channel": "shell",
        "parent_header": None,
        "content": None,
        "metadata": None,
        "header": {
            "msg_type": "msg_type_0",
            "msg_id": msg_id,
        },
    }

    components = configure(COMPONENTS, {"auth": {"mode": auth_mode}})
    async with Context() as ctx, AsyncClient():
        await JupyverseComponent(
            components=components,
            port=unused_tcp_port,
        ).start(ctx)

        # block msg_type_0
        msg["header"]["msg_id"] = str(int(msg["header"]["msg_id"]) + 1)
        kernel_server.block_messages("msg_type_0")
        async with aconnect_ws(
            f"http://127.0.0.1:{unused_tcp_port}/api/kernels/{kernel_id}/channels?session_id=session_id_0",
        ) as websocket:
            await websocket.send_json(msg)
        sleep(0.5)
        out, err = capfd.readouterr()
        assert not err

        # allow only msg_type_0
        msg["header"]["msg_id"] = str(int(msg["header"]["msg_id"]) + 1)
        kernel_server.allow_messages("msg_type_0")
        async with aconnect_ws(
            f"http://127.0.0.1:{unused_tcp_port}/api/kernels/{kernel_id}/channels?session_id=session_id_0",
        ) as websocket:
            await websocket.send_json(msg)
        sleep(0.5)
        out, err = capfd.readouterr()
        assert err.count("[IPKernelApp] WARNING | Unknown message type: 'msg_type_0'") == 1

        # block all messages
        msg["header"]["msg_id"] = str(int(msg["header"]["msg_id"]) + 1)
        kernel_server.allow_messages([])
        async with aconnect_ws(
            f"http://127.0.0.1:{unused_tcp_port}/api/kernels/{kernel_id}/channels?session_id=session_id_0",
        ) as websocket:
            await websocket.send_json(msg)
        sleep(0.5)
        out, err = capfd.readouterr()
        assert not err

        # allow all messages
        msg["header"]["msg_id"] = str(int(msg["header"]["msg_id"]) + 1)
        kernel_server.allow_messages()
        async with aconnect_ws(
            f"http://127.0.0.1:{unused_tcp_port}/api/kernels/{kernel_id}/channels?session_id=session_id_0",
        ) as websocket:
            await websocket.send_json(msg)
        sleep(0.5)
        out, err = capfd.readouterr()
        assert err.count("[IPKernelApp] WARNING | Unknown message type: 'msg_type_0'") >= 1



import psutil
from anyio import to_thread

from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User
from jupyverse_api.resource_usage import ResourceUsage, ResourceUsageConfig


class _ResourceUsage(ResourceUsage):
    def __init__(
        self,
        app: App,
        auth: Auth,
        resource_usage_config: ResourceUsageConfig,
    ):
        super().__init__(app, auth)
        self.resource_usage_config = resource_usage_config

    async def get_metrics(
        self,
        user: User,
    ):
        cur_process = psutil.Process()
        all_processes = [cur_process] + cur_process.children(recursive=True)

        # Get memory information
        rss = 0
        for p in all_processes:
            try:
                rss += p.memory_info().rss
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                pass

        mem_limit = self.resource_usage_config.mem_limit

        limits = {"memory": {"rss": mem_limit}}
        if (
            self.resource_usage_config.mem_limit
            and self.resource_usage_config.mem_warning_threshold
        ):
            limits["memory"]["warn"] = (mem_limit - rss) < (
                mem_limit * self.resource_usage_config.mem_warning_threshold
            )

        metrics = {"rss": rss, "limits": limits}

        # Optionally get CPU information
        if self.resource_usage_config.track_cpu_percent:
            cpu_count = psutil.cpu_count()
            cpu_percent = await to_thread.run_sync(_get_cpu_percent, all_processes)

            if self.resource_usage_config.cpu_limit:
                limits["cpu"] = {"cpu": self.resource_usage_config.cpu_limit}
                if self.resource_usage_config.cpu_warning_threshold:
                    limits["cpu"]["warn"] = (self.resource_usage_config.cpu_limit - cpu_percent) < (
                        self.resource_usage_config.cpu_limit
                        * self.resource_usage_config.cpu_warning_threshold
                    )

            metrics.update(cpu_percent=cpu_percent, cpu_count=cpu_count)

        return metrics


def _get_cpu_percent(all_processes):
    def get_cpu_percent(p):
        try:
            return p.cpu_percent(interval=0.05)
        # Avoid littering logs with stack traces complaining
        # about dead processes having no CPU usage
        except BaseException:
            return 0

    return sum([get_cpu_percent(p) for p in all_processes])



import asyncio
import os
from functools import partial
from pathlib import Path

import pytest
from asphalt.core import Context
from fps_yjs.ydocs import ydocs
from fps_yjs.ywebsocket import WebsocketProvider
from httpx import AsyncClient
from httpx_ws import aconnect_ws
from pycrdt import Doc, Map, Text
from utils import configure

from jupyverse_api.main import JupyverseComponent

os.environ["PYDEVD_DISABLE_FILE_VALIDATION"] = "1"

COMPONENTS = {
    "app": {"type": "app"},
    "auth": {"type": "auth", "test": True},
    "contents": {"type": "contents"},
    "frontend": {"type": "frontend"},
    "lab": {"type": "lab"},
    "jupyterlab": {"type": "jupyterlab"},
    "kernels": {"type": "kernels"},
    "yjs": {"type": "yjs"},
}


class Websocket:
    def __init__(self, websocket, roomid: str):
        self.websocket = websocket
        self.roomid = roomid

    @property
    def path(self) -> str:
        return self.roomid

    def __aiter__(self):
        return self

    async def __anext__(self) -> bytes:
        try:
            message = await self.recv()
        except Exception:
            raise StopAsyncIteration()
        return message

    async def send(self, message: bytes):
        await self.websocket.send_bytes(message)

    async def recv(self) -> bytes:
        b = await self.websocket.receive_bytes()
        return bytes(b)


@pytest.mark.asyncio
@pytest.mark.parametrize("auth_mode", ("noauth",))
async def test_execute(auth_mode, unused_tcp_port):
    url = f"http://127.0.0.1:{unused_tcp_port}"
    components = configure(COMPONENTS, {
        "auth": {"mode": auth_mode},
        "kernels": {"require_yjs": True},
    })
    async with Context() as ctx, AsyncClient() as http:
        await JupyverseComponent(
            components=components,
            port=unused_tcp_port,
        ).start(ctx)

        ws_url = url.replace("http", "ws", 1)
        name = "notebook1.ipynb"
        path = (Path("tests") / "data" / name).as_posix()
        # create a session to launch a kernel
        response = await http.post(
            f"{url}/api/sessions",
            json={
                "kernel": {"name": "python3"},
                "name": name,
                "path": path,
                "type": "notebook",
            },
        )
        r = response.json()
        kernel_id = r["kernel"]["id"]
        # get the room ID for the document
        response = await http.put(
            f"{url}/api/collaboration/session/{path}",
            json={
                "format": "json",
                "type": "notebook",
            }
        )
        file_id = response.json()["fileId"]
        document_id = f"json:notebook:{file_id}"
        ynb = ydocs["notebook"]()
        def callback(aevent, events, event):
            events.append(event)
            aevent.set()
        aevent = asyncio.Event()
        events = []
        ynb.ydoc.observe_subdocs(partial(callback, aevent, events))
        async with aconnect_ws(
            f"{ws_url}/api/collaboration/room/{document_id}"
        ) as websocket, WebsocketProvider(ynb.ydoc, Websocket(websocket, document_id)):
            # connect to the shared notebook document
            # wait for file to be loaded and Y model to be created in server and client
            await asyncio.sleep(0.5)
            # execute notebook
            for cell_idx in range(2):
                response = await http.post(
                    f"{url}/api/kernels/{kernel_id}/execute",
                    json={
                        "document_id": document_id,
                        "cell_id": ynb.ycells[cell_idx]["id"],
                    }
                )
            while True:
                await aevent.wait()
                aevent.clear()
                guid = None
                for event in events:
                    if event.added:
                        guid = event.added[0]
                if guid is not None:
                    break
            task = asyncio.create_task(connect_ywidget(ws_url, guid))
            response = await http.post(
                f"{url}/api/kernels/{kernel_id}/execute",
                json={
                    "document_id": document_id,
                    "cell_id": ynb.ycells[2]["id"],
                }
            )
            await task


async def connect_ywidget(ws_url, guid):
    ywidget_doc = Doc()
    async with aconnect_ws(
        f"{ws_url}/api/collaboration/room/ywidget:{guid}"
    ) as websocket, WebsocketProvider(ywidget_doc, Websocket(websocket, guid)):
        await asyncio.sleep(0.5)
        attrs = Map()
        model_name = Text()
        ywidget_doc["_attrs"] = attrs
        ywidget_doc["_model_name"] = model_name
        assert str(model_name) == "Switch"
        assert str(attrs) == '{"value":true}'



import uuid
from typing import Dict, List

from fastapi_users import schemas

from jupyverse_api.auth import User


class JupyterUser(User):
    anonymous: bool = True
    permissions: Dict[str, List[str]]


class UserRead(schemas.BaseUser[uuid.UUID], JupyterUser):
    pass


class UserCreate(schemas.BaseUserCreate, JupyterUser):
    pass


class UserUpdate(schemas.BaseUserUpdate, JupyterUser):
    pass



from typing import Dict

from pydantic import BaseModel

from .app import App

__version__ = "0.4.2"


class Singleton(type):
    _instances: Dict = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Config(BaseModel):
    model_config = {"extra": "forbid"}


class Router:
    _app: App

    def __init__(
        self,
        app: App,
    ) -> None:
        self._app = app

    @property
    def _type(self):
        return self.__class__.__name__

    def include_router(self, router, **kwargs):
        self._app._include_router(router, self._type, **kwargs)

    def mount(self, path: str, *args, **kwargs) -> None:
        self._app._mount(path, self._type, *args, **kwargs)

    def add_middleware(self, middleware, *args, **kwargs) -> None:
        self._app.add_middleware(middleware, *args, **kwargs)
